/*
	Copyright (C) 1995-2001 AROS - The Amiga Research OS
	$Id: inputhandler.c,v 1.11 2009/02/22 21:34:28 jacadcaps Exp $

	Desc: Intuition's InputHandler
	Lang: english
*/

/****************************************************************************************/

#define AROS_ALMOST_COMPATIBLE 1 /* NEWLIST macro */
#include <proto/exec.h>
#include <proto/intuition.h>
#include <proto/alib.h>
#include <proto/layers.h>
#include <proto/graphics.h>
#include <proto/keymap.h>
#include <proto/utility.h>
#include <proto/input.h>
#include <proto/timer.h>
#include <exec/memory.h>
#include <exec/alerts.h>
#include <exec/interrupts.h>
#include <exec/ports.h>
#include <intuition/intuition.h>
#include <intuition/intuitionbase.h>
#include <intuition/gadgetclass.h>
#include <intuition/pointerclass.h>
#include <intuition/cghooks.h>
#include <intuition/sghooks.h>
#include <devices/inputevent.h>
#include <devices/input.h>
#include <devices/rawkeycodes.h>
#include <clib/macros.h>
#include "inputhandler.h"

#include "boopsigadgets.h"
#include "boolgadgets.h"
#include "propgadgets.h"
#include "strgadgets.h"
#include "gadgets.h"
#include "intuition_intern.h" /* EWFLG_xxx */
#include "intuition_private.h"
#include "maybe_boopsi.h"
#include "inputhandler_support.h"
#include "inputhandler_actions.h"
#include "menus.h"
#include "smallmenu.h"
#include "intuition_customizesupport.h"
#include "mosmisc.h"
#include "inputhandlermos.h"

#undef DEBUG
#define DEBUG 0
#include <aros/debug.h>

#define	DEBUG_HANDLER(x)	;
#define	DEBUG_KEY(x)		;
#define	DEBUG_SCREENKEY(x)	;
#define	DEBUG_POINTER(x)	;

#ifdef __MORPHOS__
#define NOBUTTONFOCUS (IBase->IControlPrefs.ic_Flags & ICF_NOMOUSEACTIONS)
#endif

/****************************************************************************************/

struct Interrupt *InitIIH(struct IntuitionBase *IntuitionBase)
{
	struct Interrupt *iihandler;

	D(bug("InitIIH(IntuitionBase=%p)\n", IntuitionBase));

	iihandler = AllocMem(sizeof (struct Interrupt), MEMF_PUBLIC | MEMF_CLEAR);
	if (iihandler)
	{
		struct IIHData   *iihdata;

		iihdata = AllocMem(sizeof (struct IIHData), MEMF_PUBLIC | MEMF_CLEAR);
		if (iihdata)
		{
			struct MsgPort *port;

			port = AllocMem(sizeof (struct MsgPort), MEMF_PUBLIC | MEMF_CLEAR);
			if (port)
			{
				if ((iihdata->InputEventMemPool = CreatePool(MEMF_PUBLIC | MEMF_CLEAR,
								  sizeof(struct GeneratedInputEvent) * 10,
								  sizeof(struct GeneratedInputEvent) * 10)) &&
					(iihdata->ActionsMemPool = CreatePool(MEMF_SEM_PROTECTED,
								  2000, 2000)))
				{
					static const struct TagItem dragtags[] =
					{
						{GA_SysGadget	 , TRUE         },
						{GA_SysGType	, GTYP_WDRAGGING},
						{MASTERSIZE_ISDragging, TRUE    },
						{TAG_DONE			}
					};

					static const struct TagItem sizetags[] =
					{
						{GA_SysGadget	 , TRUE		    },
						{GA_SysGType	, GTYP_SIZING	},
						{TAG_DONE			}
					};

					iihdata->MasterDragGadget = (struct Gadget *)NewObjectA(IBase->sizebuttonclass,
								    NULL,
								    dragtags);

					iihdata->MasterSizeGadget = (struct Gadget *)NewObjectA(IBase->sizebuttonclass,
								    NULL,
								    sizetags);

					if (iihdata->MasterDragGadget && iihdata->MasterSizeGadget)
					{
						ULONG lock;

						/* We do not want to be woken up by message replies.
						   We are anyway woken up about 10 times a second by
						   timer events
						   */
						port->mp_Node.ln_Type = NT_MSGPORT;
						port->mp_Flags        = PA_IGNORE;

						NEWLIST( &(port->mp_MsgList) );
						iihdata->IntuiReplyPort = port;

						NEWLIST((struct List*) &iihdata->IntuiActionQueue);
						NEWLIST((struct List*)&iihdata->NewAllocatedInputEventList);
						NEWLIST((struct List*)&iihdata->AllocatedInputEventList);
						//iihdata->EndInputEventChain = &iihdata->ReturnInputEvent;
						iihdata->FreeInputEvents = NULL;

						iihdata->ActQualifier = IEQUALIFIER_RELATIVEMOUSE;

						/* Note: there are several routines like CloseWindow, which
						   expect is_Data to point to the IIHData structure, so don't
						   change this! */

						MOSIHInit(iihdata,IntuitionBase);

						iihandler->is_Code = (APTR)AROS_ASMSYMNAME(IntuiInputHandlerMOS);
						iihandler->is_Data = iihdata;
						iihandler->is_Node.ln_Pri	= 50;
						iihandler->is_Node.ln_Name	= "Intuition InputHandler";

						lock = LockIBase(0UL);

						iihdata->IntuitionBase = IntuitionBase;

						UnlockIBase(lock);

						IBase->IntuiReplyPort = iihdata->IntuiReplyPort;
						IBase->IntuiActionQueue = &iihdata->IntuiActionQueue;

						ReturnPtr ("InitIIH", struct Interrupt *, iihandler);
					} /* f (iihdata->MasterDragGadget && iihdata->MasterSizeGadget) */

					DisposeObject((Object *)iihdata->MasterDragGadget);
					DisposeObject((Object *)iihdata->MasterSizeGadget);

				} /* if (iihdata->InputEventMemPool = ... */

				if (iihdata->ActionsMemPool)
				{
					DeletePool(iihdata->ActionsMemPool);
				}
				if (iihdata->InputEventMemPool)
				{
					DeletePool(iihdata->InputEventMemPool);
				}

				FreeMem(port, sizeof(struct MsgPort));

			} /* if (port) */
			FreeMem(iihdata, sizeof (struct IIHData));

		} /* if (iihdata) */
		FreeMem(iihandler, sizeof (struct Interrupt));

	} /* if (iihandler) */

	ReturnPtr ("InitIIH", struct Interrupt *, NULL);
}

/****************************************************************************************/

VOID CleanupIIH(struct Interrupt *iihandler, struct IntuitionBase *IntuitionBase)
{
	struct IIHData *iihdata = (struct IIHData *)iihandler->is_Data;

	DisposeObject((Object *)iihdata->MasterDragGadget);
	DisposeObject((Object *)iihdata->MasterSizeGadget);

	FreeGeneratedInputEvents(iihdata);
	DeletePool(iihdata->InputEventMemPool);
	DeletePool(iihdata->ActionsMemPool);

	/* One might think that this port is still in use by the inputhandler.
	** However, if intuition is closed for the last time, there should be no
	** windows that IntuiMessage can be sent to.
	*/
	FreeMem(iihdata->IntuiReplyPort, sizeof (struct MsgPort));

	FreeMem(iihdata, sizeof (struct IIHData));
	FreeMem(iihandler, sizeof (struct Interrupt));

	return;
}

/****************************************************************************************/

// #define IDCMPDEBUG

void HandleIntuiReplyPort(struct IIHData *iihdata, struct IntuitionBase *IntuitionBase)
{
	struct IntuiMessage *im;

	while ((im = (struct IntuiMessage *)GetMsg(iihdata->IntuiReplyPort)))
	{
		if (IIM(im)->origwindow != im->IDCMPWindow && im->IDCMPWindow != NULL)
		{
			dprintf("Major bug: IntuiMessage 0x%lx was sent to window 0x%lx, but replied with IDCMPWindow set to 0x%lx!\n",
				(LONG)im,(LONG)IIM(im)->origwindow,(LONG)im->IDCMPWindow);
		}

		if (IW(im->IDCMPWindow))
		{
			struct Window *win = im->IDCMPWindow;

			Forbid();
			GetSysTime(&IW(win)->lastmsgreplied);
			Permit();

			if (IIM(im)->flags & INTUIMESSAGEF_REPEATEVENTS)
			{
				if (IW(win)->num_repeatevents)
				{
					IW(win)->num_repeatevents--;
				}
#ifdef IDCMPDEBUG
				if (!(im->Qualifier & IEQUALIFIER_REPEAT)) dprintf("REPEATEVENTS mismatch. Window %lx\n",(LONG)win);
#endif
			}

			if (IIM(im)->flags & INTUIMESSAGEF_MOUSEEVENTS)
			{
				IW(im->IDCMPWindow)->num_mouseevents--;
#ifdef IDCMPDEBUG
				if (im->Class != IDCMP_MOUSEMOVE) dprintf("MOUSEEVENTS class mismatch! Window %lx\n",(LONG)win);
#endif
			}

			if (IIM(im)->flags & INTUIMESSAGEF_INTUITICKS)
			{
				Forbid();
				im->IDCMPWindow->Flags &= ~WFLG_WINDOWTICKED;
				Permit();
#ifdef IDCMPDEBUG
				if (im->Class != IDCMP_INTUITICKS) dprintf("INTUITICKS class mismatch! Window %lx\n",(LONG)win);
#endif
			}

			if (IIM(im)->flags & INTUIMESSAGEF_IDCMPUPDATE)
			{
				if (im->IAddress)
				{
					FreeTagItems((struct TagItem *)im->IAddress);
				}

				IW(im->IDCMPWindow)->num_idcmpupdate--;

				if (!(IW(im->IDCMPWindow)->num_idcmpupdate) && IW(im->IDCMPWindow)->messagecache)
				{
					SendIntuiMessage(im->IDCMPWindow,IW(im->IDCMPWindow)->messagecache);
					IW(im->IDCMPWindow)->messagecache = 0;
				}
#ifdef IDCMPDEBUG
				if (im->Class != IDCMP_IDCMPUPDATE) dprintf("IDCMPUPDATE class mismatch! Window %lx\n",(LONG)win);
#endif
			}

			if (IIM(im)->flags & INTUIMESSAGEF_SIZEEVENTS)
			{
				IW(im->IDCMPWindow)->num_sizeevents--;
#ifdef IDCMPDEBUG
				if ((im->Class != IDCMP_NEWSIZE) && (im->Class != IDCMP_CHANGEWINDOW)) dprintf("NEWSIZE/CHANGEWINDOW class mismatch! Window %lx\n",(LONG)win);
#endif
			}

			if (IIM(im)->flags & INTUIMESSAGEF_SIZEVERIFY)
			{
				HandleIDCMP_SIZEVERIFY(im,IntuitionBase);
#ifdef IDCMPDEBUG
				if (im->Class != IDCMP_SIZEVERIFY) dprintf("SIZEVERIFY class mismatch! Window %lx\n",(LONG)win);
#endif
			}

			if (IIM(im)->flags & INTUIMESSAGEF_REFRESHVERIFY)
			{
				IW(im->IDCMPWindow)->num_refreshevents--;
			}

			if (IIM(im)->flags & INTUIMESSAGEF_REQVERIFY)
			{
			   EndRequest(win->DMRequest, win);

			   ih_fire_intuimessage(win,
					        IDCMP_REQSET,
					        0,
							win,
					        IntuitionBase);
#ifdef IDCMPDEBUG
				if (im->Class != IDCMP_REQVERIFY) dprintf("REQVERIFY class mismatch! Window %lx\n",(LONG)win);
#endif
			}

			if (IIM(im)->flags & INTUIMESSAGEF_MENUVERIFY)
			{
				struct Window *w = im->IDCMPWindow;
				struct IntScreen *scr = 0;

#ifdef IDCMPDEBUG
				if (im->Class != IDCMP_MENUVERIFY) dprintf("MENUVERIFY class mismatch! Window %lx\n",(LONG)win);
#endif
				scr = IS(w->WScreen);

				if (scr != IBase->MenuVerifyScreen ||
					scr->MenuVerifySeconds > im->Seconds ||
					(scr->MenuVerifySeconds == im->Seconds &&
					 scr->MenuVerifyMicros > im->Micros))
				{
					/* The timeout has expired, just ignore. */
				}
				else
				{
					--scr->MenuVerifyMsgCount;
					if (w == scr->MenuVerifyActiveWindow &&
						im->Code == MENUCANCEL)
					{
						ULONG lock = LockIBase(0);
						struct Window *w1;

						for (w1 = scr->Screen.FirstWindow; w1; w1 = w1->NextWindow)
						{
							if (w1->IDCMPFlags & IDCMP_MENUVERIFY && w1 != scr->MenuVerifyActiveWindow)
							{
								ih_fire_intuimessage(w1,
										     IDCMP_MOUSEBUTTONS,
										     MENUUP,
										     w1,
										     IntuitionBase);
							}
						}

						UnlockIBase(lock);

						scr->MenuVerifyActiveWindow = NULL;
						scr->MenuVerifyTimeOut = 0;
						scr->MenuVerifyMsgCount = 0;
						scr->MenuVerifySeconds = 0;
						scr->MenuVerifyMicros = 0;
						IBase->MenuVerifyScreen = NULL;
					}
					else if (scr->MenuVerifyMsgCount == 0)
					{
						struct InputEvent ie;

						/* currently we ONLY need the menu open time ! */
						ie.ie_TimeStamp.tv_secs = im->Seconds;
						ie.ie_TimeStamp.tv_micro = im->Micros;

						/* This lock will be released only when the user is
						   done with menus = when IECLASS_MENU + IESUBCLASS_MENUSTOP
						   event arrives (generated by MenuHandler task) */

						ObtainSemaphore(&IBase->MenuLock);

						if (FireMenuMessage(MMCODE_START, scr->MenuVerifyActiveWindow, &ie, IntuitionBase))
						{
							if (!IS(scr)->Compositing)
							{
								struct IOStdReq req;
								ULONG cmd = IND_SETNEWTIMER_START;

								/* we need to ask input.device to start sending us newtimer
								events as IECLASS_TIMER has too low freq for our needs */

								memclr(&req,sizeof (struct IOStdReq));

								req.io_Device = IBase->InputIO->io_Device;
								req.io_Unit = IBase->InputIO->io_Unit;
								req.io_Command = IND_SETNEWTIMER;
								req.io_Data = &cmd;
								DoIO((struct IORequest *)&req);
							}

							iihdata->MenuWindow = scr->MenuVerifyActiveWindow;
							MENUS_ACTIVE = TRUE;
						} else {
							ReleaseSemaphore(&IBase->MenuLock);
						}

						scr->MenuVerifyActiveWindow = NULL;
						scr->MenuVerifyTimeOut = 0;
						scr->MenuVerifyMsgCount = 0;
						scr->MenuVerifySeconds = 0;
						scr->MenuVerifyMicros = 0;
						IBase->MenuVerifyScreen = NULL;
					}
				}
			}
		}
		FreeIntuiMessage(im);

	} /* while ((im = (struct IntuiMessage *)GetMsg(iihdata->IntuiReplyPort))) */
}

/*******************************************************************************************************/
